'use strict';
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.compileJSFileAsBinary = void 0;
const logger_1 = require("./logger");
const node_fetch_1 = __importDefault(require("node-fetch"));
const semver_1 = __importDefault(require("semver"));
const tar_1 = __importDefault(require("tar"));
const path_1 = __importDefault(require("path"));
const zlib_1 = __importDefault(require("zlib"));
const os_1 = __importDefault(require("os"));
const rimraf_1 = __importDefault(require("rimraf"));
const crypto_1 = __importDefault(require("crypto"));
const util_1 = require("util");
const fs_1 = require("fs");
const native_addons_1 = require("./native-addons");
const executable_metadata_1 = require("./executable-metadata");
const helpers_1 = require("./helpers");
const nv_1 = __importDefault(require("@pkgjs/nv"));
async function getNodeSourceForVersion(range, dir, logger, retries = 2) {
    logger.stepStarting(`Looking for Node.js version matching ${JSON.stringify(range)}`);
    const ver = (await (0, nv_1.default)(range)).pop();
    if (!ver) {
        throw new Error(`No node version found for ${range}`);
    }
    const version = `v${ver.version}`;
    const releaseBaseUrl = `https://nodejs.org/download/release/${version}`;
    const tarballName = `node-${version}.tar.gz`;
    const cachedTarballPath = path_1.default.join(dir, tarballName);
    let hasCachedTarball = false;
    try {
        hasCachedTarball = (await fs_1.promises.stat(cachedTarballPath)).size > 0;
    }
    catch (_a) { }
    if (hasCachedTarball) {
        const shaSumsUrl = `${releaseBaseUrl}/SHASUMS256.txt`;
        logger.stepStarting(`Verifying existing tarball via ${shaSumsUrl}`);
        const [expectedSha, realSha] = await Promise.all([
            (async () => {
                try {
                    const shaSums = await (0, node_fetch_1.default)(shaSumsUrl);
                    if (!shaSums.ok)
                        return;
                    const text = await shaSums.text();
                    for (const line of text.split('\n')) {
                        if (line.trim().endsWith(tarballName)) {
                            return line.match(/^([0-9a-fA-F]+)\b/)[0];
                        }
                    }
                }
                catch (_a) { }
            })(),
            (async () => {
                const hash = crypto_1.default.createHash('sha256');
                await (0, helpers_1.pipeline)((0, fs_1.createReadStream)(cachedTarballPath), hash);
                return hash.digest('hex');
            })()
        ]);
        if (expectedSha === realSha) {
            logger.stepStarting('Unpacking existing tarball');
        }
        else {
            logger.stepFailed(new Error(`SHA256 mismatch: got ${realSha}, expected ${expectedSha}`));
            hasCachedTarball = false;
        }
    }
    let tarballStream;
    let tarballWritePromise;
    if (hasCachedTarball) {
        tarballStream = (0, fs_1.createReadStream)(cachedTarballPath);
    }
    else {
        const url = `${releaseBaseUrl}/${tarballName}`;
        logger.stepStarting(`Downloading from ${url}`);
        const tarball = await (0, node_fetch_1.default)(url);
        if (!tarball.ok) {
            throw new Error(`Could not download Node.js source tarball: ${tarball.statusText}`);
        }
        logger.stepStarting(`Unpacking tarball to ${dir}`);
        await fs_1.promises.mkdir(dir, { recursive: true });
        const contentLength = +tarball.headers.get('Content-Length');
        if (contentLength) {
            logger.startProgress(contentLength);
            let downloaded = 0;
            tarball.body.on('data', (chunk) => {
                downloaded += chunk.length;
                logger.doProgress(downloaded);
            });
        }
        tarballStream = tarball.body;
        tarballWritePromise =
            (0, helpers_1.pipeline)(tarball.body, (0, fs_1.createWriteStream)(cachedTarballPath));
    }
    try {
        await (0, helpers_1.pipeline)(tarballStream, zlib_1.default.createGunzip(), tar_1.default.x({
            cwd: dir
        }));
        await tarballWritePromise;
    }
    catch (err) {
        if (retries > 0) {
            logger.stepFailed(err);
            logger.stepStarting('Re-trying');
            return await getNodeSourceForVersion(range, dir, logger, retries - 1);
        }
        throw err;
    }
    logger.stepCompleted();
    return [version, path_1.default.join(dir, `node-${version}`)];
}
async function compileNode(sourcePath, linkedJSModules, buildArgs, makeArgs, env, logger) {
    logger.stepStarting('Compiling Node.js from source');
    const cpus = os_1.default.cpus().length;
    const options = {
        cwd: sourcePath,
        logger: logger,
        env: env
    };
    if (process.platform !== 'win32') {
        const configure = ['./configure', ...buildArgs];
        for (const module of linkedJSModules) {
            configure.push('--link-module', module);
        }
        await (0, helpers_1.spawnBuildCommand)(configure, options);
        const make = ['make', ...makeArgs];
        if (!make.some((arg) => /^-j/.test(arg))) {
            make.push(`-j${cpus}`);
        }
        if (!make.some((arg) => /^V=/.test(arg))) {
            make.push('V=');
        }
        await (0, helpers_1.spawnBuildCommand)(make, options);
        return path_1.default.join(sourcePath, 'out', 'Release', 'node');
    }
    else {
        const vcbuildArgs = [...buildArgs, ...makeArgs, 'projgen'];
        if (!vcbuildArgs.includes('debug') && !vcbuildArgs.includes('release')) {
            vcbuildArgs.push('release');
        }
        if (!vcbuildArgs.some((arg) => /^vs/.test(arg))) {
            vcbuildArgs.push('vs2019');
        }
        for (const module of linkedJSModules) {
            vcbuildArgs.push('link-module', module);
        }
        await (0, helpers_1.spawnBuildCommand)(['.\\vcbuild.bat', ...vcbuildArgs], options);
        return path_1.default.join(sourcePath, 'Release', 'node.exe');
    }
}
async function compileJSFileAsBinaryImpl(options, logger) {
    var _a, _b;
    if (!options.sourceFile.endsWith('.js')) {
        throw new Error(`Only .js files can be compiled (got: ${options.sourceFile})`);
    }
    await fs_1.promises.access(options.sourceFile);
    const namespace = options.namespace || path_1.default.basename(options.sourceFile, '.js');
    if (!options.tmpdir) {
        options.tmpdir = path_1.default.join(os_1.default.tmpdir(), 'boxednode', namespace);
    }
    const [nodeVersion, nodeSourcePath] = await getNodeSourceForVersion(options.nodeVersionRange, options.tmpdir, logger);
    const requireMappings = [];
    const extraJSSourceFiles = [];
    const enableBindingsPatch = (_a = options.enableBindingsPatch) !== null && _a !== void 0 ? _a : ((_b = options.addons) === null || _b === void 0 ? void 0 : _b.length) > 0;
    if (semver_1.default.gte(nodeVersion, '12.19.0')) {
        const extraGypDependencies = [];
        const registerFunctions = [];
        for (const addon of (options.addons || [])) {
            const addonResult = await (0, native_addons_1.modifyAddonGyp)(addon, nodeSourcePath, options.env || process.env, logger);
            for (const { linkedModuleName, targetName, registerFunction } of addonResult) {
                requireMappings.push([addon.requireRegexp, linkedModuleName]);
                extraGypDependencies.push(targetName);
                registerFunctions.push(registerFunction);
            }
        }
        logger.stepStarting('Finalizing linked addons processing');
        const nodeGypPath = path_1.default.join(nodeSourcePath, 'node.gyp');
        const nodeGyp = await (0, native_addons_1.loadGYPConfig)(nodeGypPath);
        const mainTarget = nodeGyp.targets.find((target) => ['<(node_core_target_name)', 'node'].includes(target.target_name));
        mainTarget.dependencies = [...(mainTarget.dependencies || []), ...extraGypDependencies];
        await (0, native_addons_1.storeGYPConfig)(nodeGypPath, nodeGyp);
        for (const header of ['node.h', 'node_api.h']) {
            const source = (await fs_1.promises.readFile(path_1.default.join(nodeSourcePath, 'src', header), 'utf8') +
                await fs_1.promises.readFile(path_1.default.join(__dirname, '..', 'resources', `add-${header}`), 'utf8'));
            await fs_1.promises.writeFile(path_1.default.join(nodeSourcePath, 'src', header), source);
        }
        logger.stepCompleted();
        logger.stepStarting('Handling main file source');
        let mainSource = await fs_1.promises.readFile(path_1.default.join(__dirname, '..', 'resources', 'main-template.cc'), 'utf8');
        mainSource = mainSource.replace(/\bREPLACE_WITH_ENTRY_POINT\b/g, JSON.stringify(JSON.stringify(`${namespace}/${namespace}`)));
        mainSource = mainSource.replace(/\bREPLACE_DECLARE_LINKED_MODULES\b/g, registerFunctions.map((fn) => `void ${fn}(const void**,const void**);\n`).join(''));
        mainSource = mainSource.replace(/\bREPLACE_DEFINE_LINKED_MODULES\b/g, registerFunctions.map((fn) => `${fn},`).join(''));
        await fs_1.promises.writeFile(path_1.default.join(nodeSourcePath, 'src', 'node_main.cc'), mainSource);
        logger.stepCompleted();
    }
    else {
        let tpmSource = await fs_1.promises.readFile(path_1.default.join(__dirname, '..', 'resources', 'third_party_main.js'), 'utf8');
        tpmSource = tpmSource.replace(/\bREPLACE_WITH_ENTRY_POINT\b/g, JSON.stringify(`${namespace}/${namespace}`));
        await fs_1.promises.writeFile(path_1.default.join(nodeSourcePath, 'lib', '_third_party_main.js'), tpmSource);
        extraJSSourceFiles.push('./lib/_third_party_main.js');
        let nodeCCSource = await fs_1.promises.readFile(path_1.default.join(nodeSourcePath, 'src', 'node.cc'), 'utf8');
        nodeCCSource = nodeCCSource.replace(/ProcessGlobalArgs\((?:[^{};]|[\r\n])*?kDisallowedInEnvironment(?:[^{}]|[\r\n])*?\)/, '0');
        await fs_1.promises.writeFile(path_1.default.join(nodeSourcePath, 'src', 'node.cc'), nodeCCSource);
        if (options.addons && options.addons.length > 0) {
            logger.stepStarting('Handling linked addons');
            logger.stepFailed(new Error('Addons are not supported on Node v12.x, ignoring...'));
        }
    }
    logger.stepStarting('Inserting custom code into Node.js source');
    await fs_1.promises.mkdir(path_1.default.join(nodeSourcePath, 'lib', namespace), { recursive: true });
    const source = await fs_1.promises.readFile(options.sourceFile, 'utf8');
    await fs_1.promises.writeFile(path_1.default.join(nodeSourcePath, 'lib', namespace, `${namespace}_src.js`), `module.exports = ${JSON.stringify(source)}`);
    let entryPointTrampolineSource = await fs_1.promises.readFile(path_1.default.join(__dirname, '..', 'resources', 'entry-point-trampoline.js'), 'utf8');
    entryPointTrampolineSource = entryPointTrampolineSource.replace(/\bREPLACE_WITH_BOXEDNODE_CONFIG\b/g, JSON.stringify({
        srcMod: `${namespace}/${namespace}_src`,
        requireMappings: requireMappings.map(([re, linked]) => [re.source, re.flags, linked]),
        enableBindingsPatch
    }));
    await fs_1.promises.writeFile(path_1.default.join(nodeSourcePath, 'lib', namespace, `${namespace}.js`), entryPointTrampolineSource);
    extraJSSourceFiles.push(`./lib/${namespace}/${namespace}.js`, `./lib/${namespace}/${namespace}_src.js`);
    logger.stepCompleted();
    logger.stepStarting('Storing executable metadata');
    const resPath = path_1.default.join(nodeSourcePath, 'src', 'res');
    await fs_1.promises.writeFile(path_1.default.join(resPath, 'node.rc'), await (0, executable_metadata_1.generateRCFile)(resPath, options.targetFile, options.executableMetadata));
    logger.stepCompleted();
    if (options.preCompileHook) {
        logger.stepStarting('Running pre-compile hook');
        await options.preCompileHook(nodeSourcePath, options);
        logger.stepCompleted();
    }
    const binaryPath = await compileNode(nodeSourcePath, extraJSSourceFiles, options.configureArgs, options.makeArgs, options.env || process.env, logger);
    logger.stepStarting(`Moving resulting binary to ${options.targetFile}`);
    await fs_1.promises.mkdir(path_1.default.dirname(options.targetFile), { recursive: true });
    await fs_1.promises.copyFile(binaryPath, options.targetFile);
    logger.stepCompleted();
    if (options.clean) {
        logger.stepStarting('Cleaning temporary directory');
        await (0, util_1.promisify)(rimraf_1.default)(options.tmpdir, { glob: false });
        logger.stepCompleted();
    }
}
function parseEnvVarArgList(value) {
    if (!value)
        return [];
    try {
        return JSON.parse(value);
    }
    catch (_a) {
        return value.split(',');
    }
}
async function compileJSFileAsBinary(options) {
    const logger = options.logger || new logger_1.LoggerImpl();
    const configureArgs = [...(options.configureArgs || [])];
    configureArgs.push(...parseEnvVarArgList(process.env.BOXEDNODE_CONFIGURE_ARGS));
    const makeArgs = [...(options.makeArgs || [])];
    makeArgs.push(...parseEnvVarArgList(process.env.BOXEDNODE_MAKE_ARGS));
    try {
        await compileJSFileAsBinaryImpl({
            ...options,
            configureArgs,
            makeArgs
        }, logger);
    }
    catch (err) {
        logger.stepFailed(err);
        throw err;
    }
}
exports.compileJSFileAsBinary = compileJSFileAsBinary;
//# sourceMappingURL=index.js.map