"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.modifyAddonGyp = exports.storeGYPConfig = exports.loadGYPConfig = void 0;
const crypto_1 = require("crypto");
const fs_1 = require("fs");
const gyp_parser_1 = require("gyp-parser");
const path_1 = __importDefault(require("path"));
const pkg_up_1 = __importDefault(require("pkg-up"));
const helpers_1 = require("./helpers");
async function loadGYPConfig(filename) {
    try {
        return (0, gyp_parser_1.parse)(await fs_1.promises.readFile(filename, 'utf8'));
    }
    catch (err) {
        throw new Error(`Cannot read ${filename}: ${err.message}`);
    }
}
exports.loadGYPConfig = loadGYPConfig;
async function storeGYPConfig(filename, config) {
    return await fs_1.promises.writeFile(filename, JSON.stringify(config, null, '  '));
}
exports.storeGYPConfig = storeGYPConfig;
function turnIntoStaticLibrary(config, addonId) {
    var _a, _b;
    if (!Array.isArray(config.targets))
        return [];
    const result = [];
    for (const target of config.targets) {
        if (!target.type || target.type === 'loadable_module') {
            target.type = 'static_library';
        }
        const registerFunction = `boxednode_${target.target_name}_register_${addonId}`;
        const linkedModuleName = `boxednode_${target.target_name}_${addonId}`;
        const posDefines = new Set(target['defines'] || []);
        const negDefines = new Set(target['defines!'] || []);
        for (const dontWant of [
            'USING_UV_SHARED=1', 'USING_V8_SHARED=1', 'BUILDING_NODE_EXTENSION'
        ]) {
            negDefines.add(dontWant);
            posDefines.delete(dontWant);
        }
        for (const want of [
            'BUILDING_BOXEDNODE_EXTENSION',
            `BOXEDNODE_REGISTER_FUNCTION=${registerFunction}`,
            `BOXEDNODE_MODULE_NAME=${linkedModuleName}`,
            `NAPI_CPP_CUSTOM_NAMESPACE=i${(0, crypto_1.randomBytes)(12).toString('hex')}`
        ]) {
            negDefines.delete(want);
            posDefines.add(want);
        }
        target['defines'] = [...posDefines];
        target['defines!'] = [...negDefines];
        target['win_delay_load_hook'] = 'false';
        target['dependencies'] = (_b = (_a = target['dependencies']) === null || _a === void 0 ? void 0 : _a.filter(dep => !/require\s*\(.+node-addon-api.+\)\s*\.\s*gyp/.test(dep))) !== null && _b !== void 0 ? _b : [];
        result.push({
            targetName: target.target_name,
            registerFunction,
            linkedModuleName
        });
    }
    return result;
}
async function prepForUsageWithNode(config, nodeSourcePath) {
    const nodeGypDir = path_1.default.dirname(await (0, pkg_up_1.default)({ cwd: require.resolve('node-gyp') }));
    (config.includes = config.includes || []).push(path_1.default.join(nodeGypDir, 'addon.gypi'));
    config.variables = {
        ...(config.variables || {}),
        'node_root_dir%': nodeSourcePath,
        'standalone_static_library%': '1',
        'node_engine%': 'v8',
        'node_gyp_dir%': nodeGypDir,
        'library%': 'static_library',
        'visibility%': 'default',
        'module_root_dir%': nodeSourcePath,
        'node_lib_file%': 'kernel32.lib',
        'win_delay_load_hook%': 'false'
    };
}
async function modifyAddonGyp(addon, nodeSourcePath, env, logger) {
    logger.stepStarting(`Copying addon at ${addon.path}`);
    const addonId = (0, helpers_1.objhash)(addon);
    const addonPath = path_1.default.resolve(nodeSourcePath, 'deps', addonId);
    await (0, helpers_1.copyRecursive)(addon.path, addonPath);
    logger.stepCompleted();
    await (0, helpers_1.spawnBuildCommand)([...(0, helpers_1.npm)(), 'install', '--ignore-scripts', '--production'], {
        cwd: addonPath,
        logger,
        env
    });
    logger.stepStarting(`Preparing addon at ${addon.path}`);
    const sourceGYP = path_1.default.resolve(addonPath, 'binding.gyp');
    const targetGYP = path_1.default.resolve(addonPath, '.boxednode.gyp');
    const config = await loadGYPConfig(sourceGYP);
    const addonResult = turnIntoStaticLibrary(config, addonId);
    await prepForUsageWithNode(config, nodeSourcePath);
    await storeGYPConfig(targetGYP, config);
    logger.stepCompleted();
    const targetGYPRelative = path_1.default.relative(nodeSourcePath, targetGYP);
    return addonResult.map(({ targetName, registerFunction, linkedModuleName }) => ({
        targetName: `${targetGYPRelative}:${targetName}`,
        registerFunction,
        linkedModuleName
    }));
}
exports.modifyAddonGyp = modifyAddonGyp;
//# sourceMappingURL=native-addons.js.map